@@ -1,4 +1,3 @@
-import copy
 import functools
 import json
 import logging
@@ -7,14 +6,20 @@ import traceback
 from operator import attrgetter
 
 from sen.constants import ISO_DATETIME_PARSE_STRING
-from sen.exceptions import TerminateApplication, NotifyError, NotAvailableAnymore
+from sen.exceptions import (
+    TerminateApplication, NotifyError, NotAvailableAnymore
+)
 
 import docker
 import docker.errors
 
 from sen.net import NetData
-from sen.util import calculate_cpu_percent, calculate_blkio_bytes, calculate_network_bytes, repeater, \
-    humanize_time
+from sen.util import (
+    calculate_cpu_percent, calculate_cpu_percent2, calculate_blkio_bytes,
+    calculate_network_bytes, repeater,
+    humanize_time,
+    graceful_chain_get
+)
 
 logger = logging.getLogger(__name__)
 
@@ -248,19 +253,14 @@ class DockerObject:
         return hash(self._id)
 
 
-def graceful_chain_get(d, *args):
-    if not d:
-        return None
-    t = copy.deepcopy(d)
-    for arg in args:
-        try:
-            t = t[arg]
-        except (AttributeError, KeyError, TypeError):
-            return None
-    return t
-
-
 class DockerImage(DockerObject):
+    def __init__(self, data, docker_backend, object_id=None):
+        super().__init__(data, docker_backend, object_id=object_id)
+        self._unique_size = None
+        self._total_size = None
+        self._virtual_size = None
+        self._shared_size = None
+
     @property
     def image_id(self):
         return self.object_id
@@ -372,22 +372,36 @@ class DockerImage(DockerObject):
         return self.metadata_get(["Comment"])
 
     @property
-    def size(self):
+    def total_size(self):
         """
         Size of ALL layers in bytes
 
-        :return: int
+        :return: int or None
         """
-        return self.data.get("VirtualSize", 0)
+        return self._total_size or self.data.get("Size", 0)
 
     @property
-    def layer_size(self):
+    def unique_size(self):
         """
         Size of ONLY this particular layer
 
-        :return: int
+        :return: int or None
         """
-        return self.data.get("Size", 0)
+        self._virtual_size = self._virtual_size or \
+                             graceful_chain_get(self.data, "VirtualSize", default=0)
+        try:
+            return self._virtual_size - self._shared_size
+        except TypeError:
+            return 0
+
+    @property
+    def shared_size(self):
+        """
+        I guess this is size of layers which are shared with some other image
+
+        :return: int or None
+        """
+        return self._shared_size
 
     @property
     def names(self):
@@ -443,8 +457,8 @@ class DockerImage(DockerObject):
         return self._inspect
 
     @operation("{object_type} {object_short_name} removed!")
-    def remove(self):
-        return self.d.remove_image(self.image_id)
+    def remove(self, force=False):
+        return self.d.remove_image(self.image_id, force=force)
 
     @operation("Tag of {object_type} {object_short_name} removed!")
     def remove_tag(self, tag):
@@ -501,6 +515,11 @@ class DockerContainer(DockerObject):
     Container related logic
     """
 
+    def __init__(self, data, docker_backend, object_id=None):
+        super(DockerContainer, self).__init__(data, docker_backend, object_id)
+        self.size_root_fs = None
+        self.size_rw_fs = None
+
     def __str__(self):
         return "{} ({})".format(self.container_id, self.short_name)
 
@@ -653,13 +672,23 @@ class DockerContainer(DockerObject):
 
     @operation("Get resources statistics.")
     def stats(self):
+        cpu_total = 0.0
+        cpu_system = 0.0
+        cpu_percent = 0.0
         for x in self.d.stats(self.container_id, decode=True, stream=True):
             blk_read, blk_write = calculate_blkio_bytes(x)
             net_r, net_w = calculate_network_bytes(x)
             mem_current = x["memory_stats"]["usage"]
             mem_total = x["memory_stats"]["limit"]
+
+            try:
+                cpu_percent, cpu_system, cpu_total = calculate_cpu_percent2(x, cpu_total, cpu_system)
+            except KeyError as e:
+                logger.error("error while getting new CPU stats: %r, falling back")
+                cpu_percent = calculate_cpu_percent(x)
+
             r = {
-                "cpu_percent": calculate_cpu_percent(x),
+                "cpu_percent": cpu_percent,
                 "mem_current": mem_current,
                 "mem_total": x["memory_stats"]["limit"],
                 "mem_percent": (mem_current / mem_total) * 100.0,
@@ -712,8 +741,8 @@ class DockerContainer(DockerObject):
             return logs_data
 
     @operation("{object_type} {object_short_name} removed!")
-    def remove(self):
-        self.d.remove_container(self.container_id)
+    def remove(self, force=False):
+        self.d.remove_container(self.container_id, force=force)
 
     @operation("{object_type} {object_short_name} started.")
     def start(self):
@@ -749,6 +778,7 @@ class DockerBackend:
         self._containers = None
         self._images = None  # displayed images
         self._all_images = None  # docker images -a
+        self._df = None
 
         kwargs = {"version": "auto"}
         kwargs.update(docker.utils.kwargs_from_env(assert_hostname=False))
@@ -777,6 +807,7 @@ class DockerBackend:
                 img = DockerImage(i, self)
                 self._images[img.image_id] = img
             self._all_images = {}
+            # FIXME: performance: do just all=True
             all_images_response = repeater(self.client.images, kwargs={"all": True}) or []
             for i in all_images_response:
                 img = DockerImage(i, self)
@@ -796,6 +827,30 @@ class DockerBackend:
             return [x for x in list(self._containers.values()) if x.running]
         return list(self._containers.values())
 
+    @operation("Get disk usage.")
+    def df(self, cached=True):
+        if cached is False or self._df is None:
+            logger.debug("getting disk-usage")
+            # TODO: wrap in try/execpt
+            self._df = self.client.df()
+            # TODO: attach these to real containers and images
+            # # since DOCKER API-1.25 (v.1.13.0)
+            # df = self.client.df()
+            # if 'Containers' in df:
+            #     df_containers = df['Containers']
+            containers_data = graceful_chain_get(self._df, "Containers")
+            for c_data in containers_data:
+                c = graceful_chain_get(self._containers, graceful_chain_get(c_data, "Id"))
+                c.size_root_fs = graceful_chain_get(c_data, "SizeRootFs")
+                c.size_rw_fs = graceful_chain_get(c_data, "SizeRw")
+            images_data = graceful_chain_get(self._df, "Images")
+            for i_data in images_data:
+                i = graceful_chain_get(self._images, graceful_chain_get(i_data, "Id"))
+                i._total_size = graceful_chain_get(i_data, "Size")
+                i._shared_size = graceful_chain_get(i_data, "SharedSize")
+                i._virtual_size = graceful_chain_get(i_data, "VirtualSize")
+        return self._df
+
     def realtime_updates(self):
         event = it = None
         while True:
